"""Utility functions for xdol."""
import os
import inspect


# Pattern: meshed
def resolve_module_filepath(
    module_spec, assert_output_is_existing_filepath=True
) -> str:
    if inspect.ismodule(module_spec):
        module_spec = inspect.getsourcefile(module_spec)
    elif not isinstance(module_spec, str):
        module_spec = inspect.getfile(module_spec)
    if module_spec.endswith('c'):
        module_spec = module_spec[:-1]  # remove the 'c' of '.pyc'
    if os.path.isdir(module_spec):
        module_dir = module_spec
        module_spec = os.path.join(module_dir, '__init__.py')
        assert os.path.isfile(module_spec), (
            f'You specified the module as a directory {module_dir}, '
            f"but this directory wasn't a package (it didn't have an __init__.py file)"
        )
    if assert_output_is_existing_filepath:
        assert os.path.isfile(module_spec), 'module_spec should be a file at this point'
    return module_spec


# Pattern: meshed
def resolve_to_folder(obj, assert_output_is_existing_folder=True):
    if inspect.ismodule(obj):
        obj = inspect.getsourcefile(obj)
    elif not isinstance(obj, str):
        obj = inspect.getfile(obj)

    if not os.path.isdir(obj):
        if obj.endswith('c'):
            obj = obj[:-1]  # remove the 'c' of '.pyc'
        if obj.endswith('__init__.py'):
            obj = os.path.dirname(obj)
    if assert_output_is_existing_folder:
        assert os.path.isdir(obj), 'obj should be a folder at this point'
    return obj



"""Stores for python objects """

import site
import os
from dol import wrap_kvs, filt_iter, KvReader, cached_keys, Pipe, Files
from dol.filesys import mk_relative_path_store, DirCollection, FileBytesReader
from xdol.util import resolve_to_folder


@filt_iter(filt=lambda k: k.endswith('.py') and '__pycache__' not in k)
@mk_relative_path_store(prefix_attr='rootdir')
class PyFilesBytes(FileBytesReader):
    """Mapping interface to .py files' bytes"""


# Note: One could use a more robust bytes decoder (like tec.util.decode_or_default)
bytes_decoder = lambda x: x.decode()

py_files_wrap = Pipe(
    wrap_kvs(obj_of_data=bytes_decoder),
    filt_iter(filt=lambda k: k.endswith('.py') and '__pycache__' not in k),
    mk_relative_path_store(prefix_attr='rootdir'),
)


# TODO: Extend PyFilesReader to take more kinds of src arguments.
#   for example: single .py filepaths or iterables thereof (use cached_keys for that)
# @wrap_kvs(obj_of_data=bytes_decoder)
# @filt_iter(filt=lambda k: k.endswith('.py') and '__pycache__' not in k)
# @mk_relative_path_store(prefix_attr='rootdir')
@py_files_wrap
class PyFilesReader(FileBytesReader, KvReader):
    """Mapping interface to .py files of a folder.
    Keys are relative .py paths.
    Values are the string contents of the .py file.

    Important Note: If the byte contents of the .py file can't be decoded (with a simple bytes.decode()),
    an empty string will be returned as it's value (i.e. contents).

    >>> import asyncio
    >>> s = PyFilesReader(asyncio)
    >>> assert len(s) > 10  # to test length (currently asyncio has 29 modules
    >>> 'locks.py' in s
    True

    But you can also specify an __init__.py filepath or the directory containing it.

    >>> import os
    >>> init_filepath = asyncio.__file__
    >>> dirpath_to_asyncio_modules = os.path.dirname(init_filepath)
    >>> ss = PyFilesReader(init_filepath)
    >>> sss = PyFilesReader(dirpath_to_asyncio_modules)
    >>> assert list(s) == list(ss) == list(sss)

    """

    def __init__(self, src, *, max_levels=None):
        super().__init__(rootdir=resolve_to_folder(src), max_levels=max_levels)

    def init_file_contents(self):
        """Returns the string of contents of the __init__.py file if it exists, and None if not"""
        return self.get('__init__.py', None)

    def is_pkg(self):
        """Returns True if, and only if, the root is a pkg folder (i.e. has an __init__.py file)"""
        return '__init__.py' in self


# TODO: Make it work
# @py_files_wrap
# class PyFilesText(Files):
#     def __init__(self, src, *, max_levels=None):
#         super().__init__(rootdir=resolve_to_folder(src), max_levels=max_levels)


PkgFilesReader = PyFilesReader  # back-compatibility alias

builtins_rootdir = os.path.dirname(os.__file__)
builtins_py_files = cached_keys(PyFilesReader(builtins_rootdir))

sitepackages_rootdir = next(iter(site.getsitepackages()))
sitepackages_py_files = cached_keys(PyFilesReader(sitepackages_rootdir))


@filt_iter(filt=lambda k: not k.endswith('__pycache__'))
@wrap_kvs(key_of_id=lambda x: x[:-1], id_of_key=lambda x: x + os.path.sep)
@mk_relative_path_store(prefix_attr='rootdir')
class PkgReader(DirCollection, KvReader):
    def __getitem__(self, k):
        return PyFilesReader(os.path.join(self.rootdir, k))



"""Tools for manipulating paths.

A path is a sequence of objects that represent a reference or location. 
Often paths are strings that use a special character (e.g. '/') to distinguish the 
separate objects in the sequence.

"""

import os


def extract_path_segments(path, *, sep=os.path.sep, depth=1, trailing_sep=True):
    """Extracts the first `depth` segments of a path, delimited by `sep`.

    :param path: The path to extract the segments from.
    :param sep: The delimiter to use to separate the segments.
    :param depth: The number of segments to extract.
    :param trailing_sep: Whether to add a trailing delimiter to the result.
    :return: The first `depth` segments of the path, delimited by `sep`.

    By default, the function will use depth=1 and the system's path separator as `sep`.

    >>> extract_path_segments("snap/crackle/pop/cereal", sep='/')
    'snap/'

    You can specify a different `depth`.

    >>> extract_path_segments("snap/crackle/pop/cereal", depth=2, sep='/')
    'snap/crackle/'

    Note that if the path starts with the `sep` delimiter, the result will be different.

    >>> extract_path_segments("/snap/crackle/pop/cereal", depth=2, sep='/')
    '/snap/'

    If `path` is not a string, the function will assume it's an iterable of strings,
    and return an iterable of the results.

    >>> list(extract_path_segments(
    ...     ['a.file', 'a/folder', 'a/folder_with/subfolders/', 'another/folder'],
    ...     depth=2, sep='/'
    ... ))
    ['a.file', 'a/folder', 'a/folder_with/', 'another/folder']

    You can of course specify a different delimiter.

    >>> extract_path_segments("snap.crackle.pop.cereal", sep='.', depth=2)
    'snap.crackle.'

    That last dot doesn't make sense here? Tell the function to not add the separator
    at the end of the result.

    >>> extract_path_segments(
    ...     "snap.crackle.pop.cereal", sep='.', depth=2, trailing_sep=False
    ... )
    'snap.crackle'

    """
    if not isinstance(path, str):
        from functools import partial

        paths = path  # path is actually an iterable of paths
        return map(partial(extract_path_segments, sep=sep, depth=depth), paths)
    else:
        if depth < 1:
            raise ValueError('depth must be a positive integer')

        segments = path.split(sep)

        if depth >= len(segments):
            # If the depth specified exceeds the actual number of segments in the path,
            # return the entire path as is
            return path

        # Join the segments up to the specified depth with the delimiter
        result = sep.join(segments[:depth])

        # Append a trailing delimiter to indicate the last segment
        if trailing_sep:
            result += sep

        return result
